/* $Id: buildImage.c,v 1.2 1993/10/25 23:38:33 cwikla Exp $ */

/*
/* $XConsortium: xwd.c,v 1.56 91/07/25 18:00:15 rws Exp $ */

/* Copyright 1987 Massachusetts Institute of Technology */

/*
 * xwd.c MIT Project Athena, X Window system window raster image dumper.
 *
 * This program will dump a raster image of the contents of a window into a
 * file for output on graphics printers or for other uses.
 *
 *  Author: Tony Della Fera, DEC
 *      17-Jun-85
 *
 *  Modification history:
 *
 *  11/14/86 Bill Wyatt, Smithsonian Astrophysical Observatory
 *    - Removed Z format option, changing it to an XY option. Monochrome
 *      windows will always dump in XY format. Color windows will dump
 *      in Z format by default, but can be dumped in XY format with the
 *      -xy option.
 *
 *  11/18/86 Bill Wyatt
 *    - VERSION 6 is same as version 5 for monchrome. For colors, the
 *      appropriate number of Color structs are dumped after the header,
 *      which has the number of colors (=0 for monochrome) in place of the
 *      V5 padding at the end. Up to 16-bit displays are supported. I
 *      don't yet know how 24- to 32-bit displays will be handled under
 *      the Version 11 protocol.
 *
 *  6/15/87 David Krikorian, MIT Project Athena
 *    - VERSION 7 runs under the X Version 11 servers, while the previous
 *      versions of xwd were are for X Version 10.  This version is based
 *      on xwd version 6, and should eventually have the same color
 *      abilities. (Xwd V7 has yet to be tested on a color machine, so
 *      all color-related code is commented out until color support
 *      becomes practical.)
 */

/*%
 *%    This is the format for commenting out color-related code until
 *%  color can be supported.
%*/




/* $XConsortium: xwud.c,v 1.48 91/07/25 18:02:05 rws Exp $ */
/* Copyright 1985, 1986, 1988 Massachusetts Institute of Technology */

/*
Permission to use, copy, modify, distribute, and sell this software and its
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation, and that the name of M.I.T. not be used in advertising or
publicity pertaining to distribution of the software without specific,
written prior permission.  M.I.T. makes no representations about the
suitability of this software for any purpose.  It is provided "as is"
without express or implied warranty.
*/

/* xwud - marginally useful raster image undumper */

/*
**
** Copyright (c) 1993 by Wolfram Research, Inc.  All Rights Reserved.
** 
** WOLFRAM RESEARCH, INC. ("WRI") DOES NOT WARRANT THAT THE SOFTWARE CONTAINED IN
** THIS FILE IS FREE FROM BUGS AND OMISSIONS; THE SOFTWARE IS PROVIDED AS IS.
** WRI MAKES NO REPRESENTATIONS, EXPRESS OR IMPLIED, WITH RESPECT TO THE SOFTWARE,
** INCLUDING WITHOUT LIMITATIONS ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR
** FITNESS FOR A PARTICULAR PURPOSE, ALL OF WHICH IS EXPRESSLY DISCLAIMED.  WRI
** DOES NOT WARRANT THAT THE SOFTWARE WILL MEET YOUR REQUIREMENTS OR THAT ITS
** OPERATION WILL BE UNINTERRUPTED OR ERROR FREE.
** 
** WRI DOES NOT WARRANT THAT THE SOFTWARE CONTAINED IN THIS FILE DOES NOT
** INFRINGE UPON THE INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTIES.
** 
** USE OF THE SOFTWARE CONTAINED IN THIS FILE IS ENTIRELY AT YOUR OWN RISK.
** 
** To the extent that it is authorized to do so, Wolfram Research grants the
** recipient of this software permission to use the software for his/her own
** personal purposes subject the user's acceptance of the terms stated above
** and the user's inclusion of Wolfram Reseach's copyright notice and disclaimers
** in any reproduction and/or derivative of this file.  
** 
**
**
**	John L. Cwikla
**	cwikla@wri.com
** 	Wolfram Research, Inc
*/


#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/XWDFile.h>

#include "wd.h"
#include "buildImage.h"

int DoStandardImage(Display *dpy, ColorInfo *ci, XImage *inImage, XImage *outImage, int newMapOnFail)
{
	int x, y;
	int i;
	int numberOfColors;
	XColor *xcolors, *color;
	int status;

	numberOfColors = ci->numberOfColors;
	xcolors = ci->colors;

	for(i=0;i<numberOfColors;i++)
	{
		(xcolors+i)->flags = DoRed|DoBlue|DoGreen;
		status = (ci->allocFunc)(dpy, ci->colormap, xcolors+i);
		if (status == 0)
		{
			if (newMapOnFail)
			{
				ci->colormap = XCopyColormapAndFree(dpy, ci->colormap);
				if ((*ci->allocFunc)(dpy, ci->colormap, xcolors+i) == 0)
					(xcolors+i)->pixel = ci->defaultPixel;
			}
			else
				(xcolors+i)->pixel = ci->defaultPixel;
		}
	}

	for (y = 0; y < inImage->height; y++)
	{
		for (x = 0; x < inImage->width; x++)
			XPutPixel(outImage, x, y, xcolors[XGetPixel(inImage, x, y)].pixel);
	}
}


int DoDirectImage(Display *dpy, ColorInfo *ci, XImage *inImage, XImage *outImage, int newMapOnFail)
{
	register int x, y;
	XColor color;
	unsigned long rmask, gmask, bmask;
	int rshift = 0, gshift = 0, bshift = 0;
	int i;
	unsigned long pix, xpix;
	unsigned long *pixels, *rpixels;
	int numberOfColors;
	XColor *xcolors;
	int status;

	pixels = NULL;
	rpixels = NULL;

	rmask = ci->rmask;
	while (!(rmask & 1)) 
	{
		rmask >>= 1;
		rshift++;
	}

	gmask = ci->gmask;
	while (!(gmask & 1)) 
	{
		gmask >>= 1;
		gshift++;
	}

	bmask = ci->bmask;
	while (!(bmask & 1)) 
	{
		bmask >>= 1;
		bshift++;
	}

	xcolors = ci->colors;
	numberOfColors = ci->numberOfColors;

	if (inImage->depth <= 12) 
	{
		pix = 1 << inImage->depth;
		pixels = (unsigned long *)malloc(sizeof(unsigned long) * pix);

		if (pixels == (unsigned long *)NULL)
			return 0;

		for (i = 0; i < pix; i++)
			pixels[i] = ~0L;
	
		color.flags = DoRed | DoGreen | DoBlue;
		for (y = 0; y < inImage->height; y++) 
		{
			for (x = 0; x < inImage->width; x++) 
			{
				pix = XGetPixel(inImage, x, y);
				if ((color.pixel = pixels[pix]) == ~0L) 
				{
					color.red = (pix >> rshift) & rmask;
					color.green = (pix >> gshift) & gmask;
					color.blue = (pix >> bshift) & bmask;
					if (numberOfColors) 
					{
						color.red = xcolors[color.red].red;
						color.green = xcolors[color.green].green;
						color.blue = xcolors[color.blue].blue;
					}
					else 
					{
						color.red = (((unsigned short)color.red * 65535) / rmask);
						color.green = (((unsigned short)color.green * 65535) / gmask);
						color.blue = (((unsigned short)color.blue * 65535) / bmask);
					}
					status = (ci->allocFunc)(dpy, ci->colormap, &color);
					if (status == 0)
					{
						if (newMapOnFail)
						{
								ci->colormap = XCopyColormapAndFree(dpy, ci->colormap);
								if ((*ci->allocFunc)(dpy, ci->colormap, &color) == 0)
										color.pixel = ci->defaultPixel;
						}
						else
								color.pixel = ci->defaultPixel;
					}

						
					pixels[pix] = color.pixel;
				}
				XPutPixel(outImage, x, y, color.pixel);
			}
		}
	} 
	else 
	{
		pix = 1 << 12;
		pixels = (unsigned long *)malloc(sizeof(unsigned long) * pix);

		if (pixels == (unsigned long *)NULL)
			return 0; 

		rpixels = (unsigned long *)malloc(sizeof(unsigned long) * pix);

		if (rpixels == (unsigned long *)NULL)
		{
			free(pixels);
			return 0;
		}

		for (i = 0; i < pix; i++) 
		{
			pixels[i] = ~0L;
			rpixels[i] = ~0L;
		}
	
		color.flags = DoRed | DoGreen | DoBlue;
		for (y = 0; y < inImage->height; y++) 
		{
			for (x = 0; x < inImage->width; x++) 
			{
				pix = XGetPixel(inImage, x, y);
				xpix = ((pix >> 12) ^ pix) & ((1 << 12) - 1);
				if (((color.pixel = pixels[xpix]) == ~0L) || (rpixels[xpix] != pix)) 
				{
					color.red = (pix >> rshift) & rmask;
					color.green = (pix >> gshift) & gmask;
					color.blue = (pix >> bshift) & bmask;
					if (numberOfColors) 
					{
						color.red = xcolors[color.red].red;
						color.green = xcolors[color.green].green;
						color.blue = xcolors[color.blue].blue;
					} 
					else 
					{
						color.red = (((unsigned short)color.red * 65535) / rmask);
						color.green = (((unsigned short)color.green * 65535) / gmask);
						color.blue = (((unsigned short)color.blue * 65535) / bmask);
					}
					status = (ci->allocFunc)(dpy, ci->colormap, &color);
					if (status == 0)
					{
						if (newMapOnFail)
						{
								ci->colormap = XCopyColormapAndFree(dpy, ci->colormap);
								if ((*ci->allocFunc)(dpy, ci->colormap, &color) == 0)
										color.pixel = ci->defaultPixel;
						}
						else
								color.pixel = ci->defaultPixel;
					}

					pixels[xpix] = color.pixel;
					rpixels[xpix] = pix;
				}
				XPutPixel(outImage, x, y, color.pixel);
			}
		}
	}

	if (pixels)
		free(pixels);

	if (rpixels)
		free(rpixels);

	return 1;
}


